/*****************************************************************************
 * PROJECT: TCA
 *
 * (c) Copyright Reid Simmons, 1999. All rights reserved.
 * (c) Copyright Richard Goodwin, 1995. All rights reserved.
 *
 * FILE: bridge.c
 *
 * ABSTRACT:
 * 
 * This program provides a bridge between to central servers.
 * It is based on the original bridge program by Rich Goodwin, but is extended
 * to correctly handle broadcast messages and is now a full 2-way bridge.
 *
 * $Source: /afs/cs.cmu.edu/project/TCA/Master/tcaV8/tools/bridge/bridge.c,v $ 
 * $Revision: 1.6 $
 * $Date: 99/05/09 10:54:44 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log:	bridge.c,v $
 * Revision 1.6  99/05/09  10:54:44  reids
 * New bridge program -- true 2-way bridge, and handles commands and goals
 * correctly
 * 
 * Revision 1.5  1996/02/10  16:51:34  rich
 * Made private functions static.
 *
 * Revision 1.4  1996/01/31  22:54:30  reids
 * Added automatic updating of (micro) version control numbers
 *
 * Revision 1.3  1995/12/17  20:25:07  rich
 * Minor change.
 *
 * Revision 1.2  1995/12/15  01:25:26  rich
 * Fixed the includes.
 *
 * Revision 1.1  1995/10/29  18:29:34  rich
 * Initial version of the message bridge.  It allows messages to be sent
 * from one central server to another.
 *
 *
 * 12-Sept-95 Richard Goodwin Created.
 *
 *****************************************************************************/

#include "tca/libc.h"
#include "tca.h"
#include "tca/devUtils.h"
#include "tca/tcaDev.h"
#include "tca/stdinDev.h"
#include "tca/basics.h"
#include "tca/strList.h"
#include "tca/tcaInternal.h"
#include "tca/centralMsg.h"
#include "bridge.h"
extern void tcaFree(char *item);

/* Uncomment this to trace things */
/*#define TRACE_BRIDGE*/

/*****************************************************************************
 * Global types
 *****************************************************************************/

/*****************************************************************************
 * Global variables
 *****************************************************************************/

static TCA_DEV_PTR central1_dev;
static TCA_DEV_PTR central2_dev;
static char        moduleName[MAX_STRING];
/* Need to copy the message name locally, since sending responses will
   free the TCA_REF_PTR (and thus the name will not be available) */
static char        messageName[MAX_STRING];

static STR_LIST_PTR responsesList = NULL;


/* A bit convoluted, but didn't want to mess with the tcaDev interface */
static void whichCentrals (TCA_DEV_PTR *thisCentral, 
			   TCA_DEV_PTR *otherCentral)
{
  TCA_CONTEXT_PTR currentContext, context1 = NULL;

  currentContext = tcaGetContext();
  /* Find out with device is associated with which context
     (because current interface does not allow one to access context directly).
     Fortunately, switching contexts is very fast, */
  if (central1_dev) {
    TCA_setContext(central1_dev);
    context1 = tcaGetContext();
  }

  if (currentContext == context1) {
    *thisCentral = central1_dev; *otherCentral = central2_dev;
  } else {
    *thisCentral = central2_dev; *otherCentral = central1_dev;
  }

  if (*otherCentral && !TCA_isConnected(*otherCentral)) *otherCentral = NULL;
  /* Restore the current context */
  tcaSetContext(currentContext);
}

/*****************************************************************************
 *
 * FUNCTION: void forwardingTaskTreeHandler(TCA_REF_PTR ref, char *data,
 *                                          BOOLEAN isCommandP)
 *
 * DESCRIPTION: Handles both commands and goals
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/
static void forwardingTaskTreeHandler(TCA_REF_PTR ref, char *data,
				      BOOLEAN isCommandP)
{
  TCA_DEV_PTR thisCentral, otherCentral;
  TCA_RETURN_VALUE_TYPE retVal;

  strncpy(messageName, tcaReferenceName(ref), MAX_STRING);
  whichCentrals(&thisCentral, &otherCentral);
  if (otherCentral != NULL) {
    TCA_setContext(otherCentral);
    if (tcaMessageHandlerRegistered(messageName)) {
      retVal = (isCommandP
		? tcaWaitForCommandWithConstraints(NULL, messageName,
						   data, NO_TPLCONSTR)
		: tcaWaitForGoalWithConstraints(NULL, messageName,
						data, NO_TPLCONSTR));
      TCA_setContext(thisCentral);
      if (retVal == Success) 
	tcaSuccess(ref);
      else
	tcaFailure(ref, (isCommandP ? "Command Failed" : "Goal Failed"), NULL);
    } else {
      TCA_setContext(thisCentral);
      tcaFailure(ref, "No Handler", NULL);
    }
  } else {
    tcaFailure(ref, "No Connection", NULL);
  }

  tcaFreeData(messageName, data);
}

/*****************************************************************************
 *
 * FUNCTION: void forwardingCommandHandler(TCA_REF_PTR ref, char *data)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void forwardingCommandHandler(TCA_REF_PTR ref, char *data)
{
  forwardingTaskTreeHandler(ref, data, TRUE);
}

/*****************************************************************************
 *
 * FUNCTION: void forwardingGoalHandler(TCA_REF_PTR ref, char *data)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void forwardingGoalHandler(TCA_REF_PTR ref, char *data)
{
  forwardingTaskTreeHandler(ref, data, FALSE);
}

/*****************************************************************************
 *
 * FUNCTION: void forwardingBroadcastHandler(TCA_REF_PTR ref, char *data)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void forwardingBroadcastHandler(TCA_REF_PTR ref, char *data)
{
  TCA_DEV_PTR thisCentral, otherCentral;

  strncpy(messageName, tcaReferenceName(ref), MAX_STRING);
  whichCentrals(&thisCentral, &otherCentral);
  if (otherCentral != NULL) {
    TCA_setContext(otherCentral);
    tcaBroadcast(messageName, data);
  }
  tcaFreeData(messageName, data);
}

/*****************************************************************************
 *
 * FUNCTION: void forwardingInformHandler(TCA_REF_PTR ref, char *data)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void forwardingInformHandler(TCA_REF_PTR ref, char *data)
{
  TCA_DEV_PTR thisCentral, otherCentral;

  strncpy(messageName, tcaReferenceName(ref), MAX_STRING);
  whichCentrals(&thisCentral, &otherCentral);
  if (otherCentral != NULL) {
    TCA_setContext(otherCentral);
    tcaInform(messageName, data);
  }
  tcaFreeData(messageName, data);
}

/*****************************************************************************
 *
 * FUNCTION: void forwardingQueryHandler(TCA_REF_PTR ref, char *data)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void forwardingQueryHandler(TCA_REF_PTR ref, char *data)
{
  TCA_RETURN_VALUE_TYPE result;
  TCA_DEV_PTR thisCentral, otherCentral;
  void *reply=NULL;

  strncpy(messageName, tcaReferenceName(ref), MAX_STRING);
  whichCentrals(&thisCentral, &otherCentral);
  if (otherCentral != NULL) {
    reply = tcaAllocateReply(messageName);
    TCA_setContext(otherCentral);
    if (tcaMessageHandlerRegistered(messageName)) {
      result = tcaQuery(messageName, data, reply);
      TCA_setContext(thisCentral);
      tcaReply(ref, reply);
    } else {
      TCA_setContext(thisCentral);
      tcaNullReply(ref);
    }
  } else {
    tcaNullReply(ref);
  }

  tcaFreeData(messageName, data);
  tcaFreeReply(messageName, reply);
  tcaFree(reply);
}

static void registerRemoteMsg (TCA_DEV_PTR otherCentral,
			       const char *name, TCA_MSG_CLASS_TYPE msg_class,
			       const char *msgFormat, const char *resFormat)
{
  TCA_setContext(otherCentral);
  if (!tcaMessageRegistered(name)) {
#ifdef TRACE_BRIDGE
    fprintf(stderr, "registerRemoteMsg from %d: %s %d\n", 
	    (otherCentral == central2_dev ? 1 : 2), name, msg_class);
#endif
    switch (msg_class) {
    case InformClass    : tcaRegisterInformMessage(name, msgFormat); break;
    case BroadcastClass : tcaRegisterBroadcastMessage(name, msgFormat); break;
    case GoalClass      : tcaRegisterGoalMessage(name, msgFormat); break;
    case CommandClass   : tcaRegisterCommandMessage(name, msgFormat); break;
    case QueryClass :
      tcaRegisterQueryMessage(name, msgFormat, resFormat); break;
    default : fprintf(stderr, "WARNING: Cannot handle message %s of type %d\n",
		      name, (int)msg_class); break;
    }
  }
}

static void registerForwardingHandler (TCA_DEV_PTR thisCentral,
				       TCA_DEV_PTR otherCentral,
				       const char *msgName, 
				       const char *hndName,
				       TCA_MSG_CLASS_TYPE msg_class)
{
  MSG_INFO_TYPE msgInfo;
  TCA_HND_FN    handler = NULL;

  TCA_setContext(otherCentral);
  msgInfo.name = (char *)msgName;
  tcaGetMsgInfo(&msgInfo);
  if (msgInfo.numberOfHandlers == 0 || msg_class == BroadcastClass) {
#ifdef TRACE_BRIDGE
    fprintf(stderr, "registerForwardingHandler from %d: %s\n", 
	    (otherCentral == central2_dev ? 1 : 2), msgName);
#endif
    switch (msg_class) {
    case QueryClass     : handler = forwardingQueryHandler; break;
    case CommandClass   : handler = forwardingCommandHandler; break;
    case InformClass    : handler = forwardingInformHandler; break;
    case BroadcastClass : handler = forwardingBroadcastHandler; break;
    case GoalClass      : handler = forwardingGoalHandler; break;
    default : fprintf(stderr, "WARNING: Cannot handle message %s of type %d\n",
		      msgName, (int)msg_class); break;
    }
    tcaRegisterHandler(msgName, hndName, handler);
  } else if (strcmp(msgInfo.resourceName, moduleName)) {
    fprintf(stderr,
	    "WARNING: Handler for %s already registered in other central\n",
	    msgName);
  }
}

/*****************************************************************************
 *
 * FUNCTION: void newMsgHnd(TCA_REF_PTR ref, MSG_REG_DATA_PTR msgRegData)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void newMsgHnd(TCA_REF_PTR ref, MSG_REG_DATA_PTR msgRegData)
{
  TCA_DEV_PTR thisCentral, otherCentral;

  whichCentrals(&thisCentral, &otherCentral);
#ifdef TRACE_BRIDGE
  fprintf(stderr, "newMsgHnd from %d: %s\n", 
	  (otherCentral == central2_dev ? 1 : 2), msgRegData->name);
#endif
  if (otherCentral != NULL) {
    registerRemoteMsg(otherCentral, msgRegData->name, msgRegData->msg_class,
		      msgRegData->msgFormat, msgRegData->resFormat);
  }
}

/*****************************************************************************
 *
 * FUNCTION: void newHndHnd(TCA_REF_PTR ref, MSG_REG_DATA_PTR msgRegData)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void newHndHnd(TCA_REF_PTR ref, HND_DATA_PTR hndData)
{
  MSG_INFO_TYPE msgInfo;
  TCA_DEV_PTR thisCentral, otherCentral;

  whichCentrals(&thisCentral, &otherCentral);
  if (otherCentral != NULL) {
#ifdef TRACE_BRIDGE
    fprintf(stderr, "newHndHnd from %d: %s\n", 
	    (otherCentral == central2_dev ? 1 : 2), hndData->msgName);
#endif
    TCA_setContext(thisCentral);
    msgInfo.name = (char *)hndData->msgName;
    tcaGetMsgInfo(&msgInfo);
    /* Don't register for handler if it is from the bridge itself */
    if ((msgInfo.numberOfHandlers == 0 ||
	 strcmp(msgInfo.resourceName, moduleName)) &&
	/* Need to handle this specially */
	strcmp(msgInfo.name, "V_watch_TCA_BROADCAST_MSG_VAR")) {
      registerForwardingHandler(thisCentral, otherCentral, msgInfo.name,
				hndData->hndName, msgInfo.msg_class);
    }
  }
}

/*****************************************************************************
 *
 * FUNCTION: void registerAll(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void registerAll(void)
{
  /* Need to set up taps to get new messages and new handlers as they are
     registered. */
  
  tcaRegisterInform(NEW_MESSAGE_LISTENER,TCA_REGISTER_MSG_INFORM_FORMAT,
		    newMsgHnd);
  tcaTapMessage(WhenSent,TCA_REGISTER_MSG_INFORM, NEW_MESSAGE_LISTENER);

  tcaRegisterInform(NEW_HANDLER_LISTENER,TCA_REGISTER_HND_INFORM_FORMAT,
		    newHndHnd);
  tcaTapMessage(WhenSent,TCA_REGISTER_HND_INFORM, NEW_HANDLER_LISTENER);
}

/*****************************************************************************
 *
 * FUNCTION: static void crossRegisterMsgs(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void crossRegisterMsgs (TCA_DEV_PTR thisCentral, 
			       TCA_DEV_PTR otherCentral)
{
  char **msgNames, **tmpNames;
  MSG_INFO_TYPE msgInfo;
  
  if (thisCentral != NULL && otherCentral != NULL &&
      TCA_isConnected(thisCentral) && TCA_isConnected(otherCentral)) {
    TCA_setContext(thisCentral);
    tmpNames = msgNames = tcaGetRegisteredMsgs();
    while (*msgNames != NULL) {
      TCA_setContext(thisCentral);
      msgInfo.name = *msgNames;
      tcaGetMsgInfo(&msgInfo);
      /* Don't register for messages for the bridge itself */
      if (strcmp(msgInfo.resourceName, moduleName)) {
	registerRemoteMsg(otherCentral, msgInfo.name, msgInfo.msg_class, 
			  msgInfo.msgFormat, msgInfo.resFormat);
	registerForwardingHandler(thisCentral, otherCentral, msgInfo.name,
				  msgInfo.name, msgInfo.msg_class);
      }
      tcaFree(*msgNames);
      msgNames++;
    }
    tcaFree((char *)tmpNames);
  }
}

/*****************************************************************************
 *
 * FUNCTION: void disconnectTCA(void)
 *
 * DESCRIPTION: 
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void disconnectTCA(void)
{
  /* called when a central disconnect is detected. */
}

/*****************************************************************************
 *
 * FUNCTION: static void reconnectTCA(void)
 *
 * DESCRIPTION: Called when a central is reconnected.
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 *****************************************************************************/

static void reconnectTCA(void)
{
  TCA_DEV_PTR thisCentral, otherCentral;

  whichCentrals(&thisCentral, &otherCentral);
  crossRegisterMsgs(thisCentral, otherCentral);
  crossRegisterMsgs(otherCentral, thisCentral);
}

/*****************************************************************************
 *
 * FUNCTION: void displayVersion()
 *
 * DESCRIPTION: Displays the version information for the program
 *
 * INPUTS: none.
 *
 * OUTPUTS: void
 *
 *****************************************************************************/

static void displayVersion(void)
{
  printf("TCA Bridge %d.%d.%d \n",
	 BRIDGE_VERSION_MAJOR, BRIDGE_VERSION_MINOR, BRIDGE_VERSION_MICRO);
  printf(" Released : %s\n", BRIDGE_VERSION_DATE);
  printf(" Commited : %s\n", BRIDGE_COMMIT_DATE);
  printf(" Compiled : %s %s\n", __DATE__, __TIME__);
  fflush(stdout);
}

int main(int argc, char **argv)
{ 
  if (argc < 2) {
    printf("Must supply name of second central server machine.\n");
    exit(-1);
  }
  
  displayVersion();

  sprintf(moduleName,"%s%s",BRIDGE_NAME_PREFIX,argv[1]);
  central1_dev = TCA_connect(moduleName, registerAll, 
			     disconnectTCA, reconnectTCA, NULL, NULL);
  if (central1_dev) {
    printf("Connected to host %s\n", tcaServerMachine());
  }
  
  TCA_setCentralHost(argv[1]);
  central2_dev = TCA_connect(moduleName, registerAll,
			   disconnectTCA, reconnectTCA, NULL, NULL);
  if (central2_dev) {
    printf("Connected to host %s\n", argv[1]);
  }
  
  responsesList = strListCreate();

  /* May want to limit the crossed registered messages in some way.  
   * For example by providing a list of acceptable messages in a startup 
   * file.
   */

  crossRegisterMsgs(central1_dev, central2_dev);
  crossRegisterMsgs(central2_dev, central1_dev);
  
  /* connect up the handlers for standard in. */
  stdin_connect(stdin_defaultInputHnd); 
  
  devMainLoop();
  exit(0);
}
