/*****************************************************************************
 * PROJECT: Xavier
 *
 * (c) Copyright 1993 Richard Goodwin & Joseph O'Sullivan. All rights reserved.
 *
 * FILE: runConsole.c
 *
 * ABSTRACT: This is the common code between xRunConsole and xfRunConsole
 *
 * $Source: /afs/cs.cmu.edu/project/TCA/Master/tcaV8/tools/supervise/inspect.c,v $
 * $Revision: 1.4 $
 * $Date: 97/05/26 19:55:48 $
 * $Author: josullvn $
 *
 * REVISION HISTORY:
 *
 * $Log:	inspect.c,v $
 * Revision 1.4  97/05/26  19:55:48  josullvn
 * Running inspect without supervise would cause inspect to crash, due to
 * jl moving the syncwithsupervise to early in the mainline. Added a flag
 * x11Running, to protect this condition.
 * 
 * Revision 1.3  1997/05/22 08:16:28  josullvn
 * Mods from jl@cs - remove superviseCatchChild race condition, inspect now
 * immediately connects with supervise, and improve the checking of running
 * processes at startup
 *
 * Revision 1.2  1997/05/20 05:19:16  josullvn
 * new flag -silent for silent operation, better message when readline fails,
 * cleaned up readlines linktime list, quit in inspect closes xterms properly
 *
 * Revision 1.1  1997/05/18 01:31:54  josullvn
 * Reverted back to the old format for storing the sources, ie
 * not in a directory hierarchy. Also updated the help system in inspect.
 *
 * Revision 1.1  1997/05/13 18:01:23  josullvn
 * Amended nanny and xfMiniRunConsole to supervise and inspect.
 * Numerous bugs fixes, but chiefly...
 *  Updated the xforms display.
 *  respond correctly to CTRL-C's.
 *  new resource file format partially available as -oldload foo.rc
 *  fixed signal handling - a change to devUtils disables prior interrupt handlers.
 * 			Not a good thing methinks...
 *  fixed fd handling. Now superviseCatchChild explicitly handles closed fd's.
 *  reduced message traffic, pings only sent when necessary
 *
 * inspect has been cleaned up, slimmed down. Communication and
 * resource management stuff in common. Still a few TODO things...
 *  Doesn't handle blank key strokes (ie just hitting carriage return).
 *  VT100 compliance is limited. As in, still can't play hangman or do vi, or
 *  anything that uses the secondary graphics window. Don't know why really, look
 *  at.... http://sv1pc44.cs.york.ac.uk/vt100_2.html
 *  Basic curses should be OK
 *
 * Revision 1.11  1996/06/28 14:07:34  reids
 * Fixed quite a few bugs -- with graphics, interaction with script, and
 *   killing processes
 *
 * Revision 1.10  1996/03/29  15:57:23  reids
 * Consolidated the common code between xCallbacks and xfCallbacks, and
 *   xRunConsole and xfRunConsole.
 * Added a way to add macro definitions to resource files ("define: <x> <y>").
 * Fixed a bug that was causing the xfRunConsole to crash when a task was
 *   restarted.
 *
 *****************************************************************************/
#include "tca/libc.h"
#include "tca/devUtils.h"

#include "messages.h"
#include "communication.h"

#include "supervise.h" /* For the version number */

#include "inspect.h"
#include "callbacks.h"
#include "script.h"

int nConsoleProcs = 0;
int currentProc   = -1;
consoleProcType consoleProc[MAX_PROCESSES];
int silent        = 0;
int x11Running    = 0;

static char *scriptFile = NULL;
int dumpFd = -1;

static char buf[DEFAULT_LINE_LENGTH];

void runNewProcess (char *process)
{
  messageMakeExec(buf, process, getenv("DISPLAY"));
  if (safePassLocalMessage(buf)) {
    sprintf(buf,"Waiting for %s...", process);
    activateStatus(buf);
    sprintf(buf, "[CONSOLE] %s waiting for confirmation from supervise\n", 
	    process);
    MessagePrintf(buf);
  }
}

void runKillProcess (char *process) 
{ 
  messageMakeKill(buf, process);
  if (safePassLocalMessage(buf)) {
    sprintf(buf, "[CONSOLE] %s waiting for supervise kill confirmation\n",  
	    process);
    MessagePrintf(buf);
  }
}

void runSuspendProcess (char *process)
{
  messageMakePause(buf, process);
  if (safePassLocalMessage(buf)) {
    sprintf(buf, "[CONSOLE] %s suspended.\n", process);
    MessagePrintf(buf);
  }
}

void runContinueProcess (char *process)
{
  messageMakeBreak(buf, process);
  if (safePassLocalMessage(buf)) {
    sprintf(buf, "[CONSOLE] continuing %s\n", process);
    MessagePrintf(buf);
  }
}

void runCommandProcess (char *process, char *command)
{
  messageMakeCommand(buf, process, command);
  if (!safePassLocalMessage(buf)) {
    sprintf(buf, "[CONSOLE] cannot send to %s\n", process);
    MessagePrintf(buf);
  }
}

static void displayVersion (void)
{
  if (silent) return;

  fprintf(stderr, "inspect %d.%d.%d \n",
	  SUPERVISE_VERSION_MAJOR, SUPERVISE_VERSION_MINOR, SUPERVISE_VERSION_MICRO);
  fprintf(stderr, " Released : %s\n", SUPERVISE_VERSION_DATE);
  fprintf(stderr, " Commited : %s\n", SUPERVISE_COMMIT_DATE);
  fprintf(stderr, " Compiled : %s %s\n", __DATE__, __TIME__);
}

static void printHelp (void)
{
  printf("Usage: inspect [options]\n");
  printf("Without options, displays a Xwindow allowing for GUI interaction.\n");
  printf("With options, allows for interaction with the Xavier supervise daemons:\n");
  printf(" -h | -help:   Print this message\n");
  printf(" -v        :   Display version information\n");
  printf(" -dump name :  Dumps all output that goes to the Xterm to a logfile\n");
  printf(" -run <name>:  Runs program name (eg -run CTR)\n");
  printf(" -kill <name>: Kills program name, if it is running\n");
  printf(" -reload [current | <file>]: Forces supervises to reload either the\n");
  printf("               current resource file or from a new location\n");
  printf(" -silent:      no command prompt/messages while inspect is running\n");
  printf("               (important for when inspect is to be backgrounded)\n");
  printf("               Note that NO warning messages will now be available\n");
  printf(" -script <file>: Load and run the given script file\n");
}

static void parseCommandLineOptions (int argc, char **argv)
{
  char *name = NULL, *file = NULL;
  char buffer[DEFAULT_LINE_LENGTH];
  int  i;

  for (i=1; i<argc && argv[i][0]=='-'; i++) {
    if (!strcmp(argv[i],"-run") && i+1<argc) {
      name = (argv[++i]);
      messageMakeExec(buffer, name, getenv("DISPLAY"));
      passLocalMessage(buffer);
      sleep(3);
      exit (0);
    } else if (!strcmp(argv[i],"-kill") && i+1<argc) {
      name = (argv[++i]);
      messageMakeKill(buffer, name);
      passLocalMessage(buffer);
      sleep(3);
      exit (0);
    } else if (!strcmp(argv[i], "-reload") && i+1<argc) {
      file = (argv[++i]);
      messageMakeReload(buffer);
      passLocalMessage(buffer);
      sleep(3);
      exit (0);
    } else if (!strcmp(argv[i], "-dump") && i+1<argc) {
      if (dumpFd == -1) close(dumpFd);
      dumpFd = open(argv[++i],O_CREAT | O_RDWR);
    } else if (!strcmp(argv[i], "-silent")) {
      silent = TRUE;
    } else if (!strcmp(argv[i], "-script") && i+1<argc) {
      scriptFile = (argv[++i]);
    } else if ((!strcmp(argv[i],"-help")) || (!strcmp(argv[i], "-h"))) {
      printHelp();
      exit(1);
    } else if (!strcmp(argv[i], "-v")) {
      displayVersion();
      exit(1);
    } else {
      WARNING2("Unknown option: %s\n", argv[i]);
      printHelp();
      exit(-1);
    }
  }
}

/* DESCRIPTION: Handles character typed from stdin.
 *
 * INPUTS:
 *
 * OUTPUTS:
 */
static void stdin_inputHnd(int fd, long chars_available)
{
  static char buffer[DEFAULT_LINE_LENGTH+1];
  int numRead=0;
  char *buffPtr;
  
  bzero(buffer, DEFAULT_LINE_LENGTH+1);
  
  numRead = devReadN(fd, buffer,
		     MIN(chars_available,DEFAULT_LINE_LENGTH));
  for(buffPtr = buffer; ((*buffPtr == ' ') && (buffPtr != '\0')); buffPtr++);
}

static void processLocalMessage(int id, char *message)
{
  messageParseLine(TRUE, NULL, id, message);
}

void connectWithSupervise (void)
{
  supervise_connect(processLocalMessage);
  
  /* Announce that we be here */
  syncWithSupervise();
}

static void initConsoleProc (void)
{
  int i;

  for (i=0; i<MAX_PROCESSES; i++) {
    consoleProc[i].graphics = createConsoleGraphics();
  }
}

void main(int argc, char **argv)
{

  /* Have to connect before some of the command line options can be
   * assimilated (e.g, -run, -kill)
   * jl@cs
   */
  connectWithSupervise();

  parseCommandLineOptions(argc, argv);

  displayVersion();

  /* Need (some of) this to initialize the list package */
  {
    extern void tcaModuleInitialize(void);
    tcaModuleInitialize();
  }

  if (!silent) {
    fprintf(stdout, ">");
    fflush(stdout);
  }

  initConsoleProc();

  /* Want sigint handled here, not by the bozo lower layers, thus 
   * make devInit explicit */
  devInit();

  xRunConsole(argc, argv);
  x11Running = TRUE;

  /* Preload a script file */
  if (scriptFile) {
    if (!scriptLoad(scriptFile)) {
      fprintf(stderr, "Cannot load script file %s\n", scriptFile);
    }
  }

  /* loop forever in the main device loop - don't wan't another devInit call... */
  for(;;)
    devProcessDevices(NULL);
  
  exit(0);
}
