/* willow.c */
/* Willow -- A tree layout package for Tcl/Tk */

#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <string.h>
#include "willow.h"

#define PROGRAM_DEBUG 1

#define MAX(a,b) ((a > b) ? a : b)

/* --------------------- Local Prototypes ---------------------------- */

void Display_Tree_Iter(Willow_Tree_Type t, Willow_Node node,
		       Willow_Node parent, int offset1, int offset2,
		       int depth);
void Willow_Move_Node(Willow_Tree_Type t, Willow_Node n,
		      int delta_x, int delta_y);
int Node_Bounding_Dimension(Willow_Tree_Type t, Willow_Node n, int depth);
int Subtree_Dimension(Willow_Tree_Type t, Willow_Node parent, int depth);
void Willow_Draw_Connector(Willow_Tree_Type t, Willow_Node n,
			   Willow_Node parent, int dx, int dy);
int Willow_Position_Tree_Nodes(Willow_Tree_Type t);
int Willow_Display_Tree(Willow_Tree_Type t);
int Willow_Tree_Depth(Willow_Tree_Type t, int max_depth);
void Willow_Redraw_Tree_Iter(Willow_Tree_Type t, Willow_Node n);
void Willow_Node_Callback(Willow_Tree_Type t, Willow_Node n);

/* ---------------------- Local Variables ---------------------------- */


/* ---------------------- Tree Operations ---------------------------- */

/* Willow_Create_Tree initializes the data structures used by the
   tree. The interp argument is the Tcl interpreter being used, canvas
   is the canvas widget that should be used for displaying the
   tree. */

void Willow_Create_Tree(Tcl_Interp *interp, char *canvas_name, 
			int orientation, int x_offset, int y_offset, 
			Willow_Tree_Type *tree)
{
  (*tree) = (Willow_Tree_Type)malloc(sizeof(struct Willow_Tree));
  
  (*tree)->interp = interp;
  (*tree)->orientation = orientation;
  (*tree)->canvas_name = canvas_name;
  (*tree)->root_node = NULL;
  (*tree)->canvas_x = x_offset;
  (*tree)->canvas_y = y_offset;
  (*tree)->callback_language = -1;
  (*tree)->child_space = CHILD_SPACING;  
  (*tree)->sibling_space = SIBLING_SPACING;  
}

/* the return result is the width of the tree */

int Willow_Position_Tree_Nodes(Willow_Tree_Type t)
{
  int tree_width;
  int level;

  /* Keep track of the maximum depth of any given layer of the tree */
  for (level = 0; level < MAX_TREE_DEPTH; level++) {
    t->max_level_size[level] = 0;
  }

  /* Compute the relative positions for all the nodes in the tree */
  tree_width = Node_Bounding_Dimension(t, t->root_node, 0);

  return tree_width;
}

int Willow_Display_Tree(Willow_Tree_Type t)
{
  /* max_depth is updated within Display_Tree_Iter */
  t->max_depth = 0;

  /* Traverse the entire tree laying out each node of the tree */
  Display_Tree_Iter(t, t->root_node, NULL, 0, 0, 0);

  return (Willow_Tree_Depth(t, t->max_depth));
}

int Willow_Tree_Depth(Willow_Tree_Type t, int max_depth)
{
  if (t->orientation == WILLOW_VERTICAL) {
    return (max_depth - t->canvas_y);
  } else {
    return (max_depth - t->canvas_x);
  }
}

void Willow_Draw_Tree(Willow_Tree_Type t, int *x, int *y)
{
  *x = Willow_Position_Tree_Nodes(t);
  *y = Willow_Display_Tree(t);
}

/* ----------- Functions dealing with a redraw of the tree ---------- */

void Willow_Set_Tcl_Callback(Willow_Tree_Type tree, char *fn)
{
  strcpy(tree->tcl_callback_fn, fn);
  tree->callback_language = 0; /* Tcl */
}

void Willow_Set_C_Callback(Willow_Tree_Type tree, 
			   void (* c_callback_fn)(void *))
{
  tree->c_callback_fn = c_callback_fn;
  tree->callback_language = 1; /* C */
}

/* this is used when you make a change to one of the global variables
   over a tree, (such as the distance between nodes, or the size of a
   font) which requires that the entire geometry of the tree be
   recomputed. */

void Willow_Redraw_Tree(Willow_Tree_Type t)
{
  Willow_Redraw_Tree_Iter(t, t->root_node);
}

/* Willow_Redraw_Tree_Iter should only be called by Willow_Redraw_Tree
   (and recursively). It goes through the tree calling the redraw
   callback function for every node of the tree. */

void Willow_Redraw_Tree_Iter(Willow_Tree_Type t, Willow_Node n)
{
  if (n == NULL) return;

  while (n) {
    Willow_Node_Callback(t, n);
    Willow_Redraw_Tree_Iter(t, n->children);
    n = n->sibling;
  }
}

/* Who deletes the node? Willow or package user? */

void Willow_Node_Callback(Willow_Tree_Type t, Willow_Node n)
{
  char tcl_command[200];

  switch (t->callback_language) {
  case 0:
    /* Callback language is Tcl */
    sprintf(tcl_command, "%s %d", t->tcl_callback_fn,
	    (int)n->additional);
    Tcl_Eval(t->interp, tcl_command);
    break;
  case 1:
    /* Callback language is C */
    (t->c_callback_fn)(n->additional);
    break;
  default:
    /* no callback language specified */
    fprintf(stderr, "Willow : No callback function specified\n");
    break;
  }
}

void Willow_Destroy_Tree(Willow_Tree_Type t)
{
  /* need to reclaim all allocated space and delete any remaining
     nodes */
}

/* ------------------------------------------------------------------- */

void Display_Tree_Iter(Willow_Tree_Type t, Willow_Node node,
		       Willow_Node parent, int offset1, int offset2,
		       int depth)
{
  int x, y;
  int delta_x, delta_y;
  int child_ofs;

  /* determine the new x,y position for the given node */
  if (t->orientation == WILLOW_VERTICAL) {
    x = t->canvas_x + (node->max / 2) - (node->width / 2) + offset1;
    y = t->canvas_y + offset2;
    t->max_depth = MAX(y + node->height, t->max_depth);
  } else {
    x = t->canvas_x + offset2;
    y = t->canvas_y + (node->max / 2) - (node->height / 2) + offset1;
    t->max_depth = MAX(x + node->width, t->max_depth);
  }

  /* if the parent is larger this provides appropriate offset */
  child_ofs = (node->max - node->children_dimension) / 2;

  /* how much does node have to be moved from it's current position? */
  delta_x = (x - node->current_x);
  delta_y = (y - node->current_y);

  if ((delta_x != 0) || (delta_y != 0)) {
    /* Need to move the given node */
    Willow_Move_Node(t, node, delta_x, delta_y);
  }

  /* with a little extra record keeping (to determine whether or not
     the parent has moved, this call does not have to made every
     iteration. */
  Willow_Draw_Connector(t, node, parent, delta_x, delta_y);

  /* now layout the children of this node */
  if (node->children) {
    Display_Tree_Iter(t, node->children, node,
		      (offset1 + child_ofs),
		      (offset2 + t->max_level_size[depth] + t->child_space),
		      (depth + 1));
  }
  
  /* and then layout the siblings of this node */
  if (node->sibling) {
    Display_Tree_Iter(t, node->sibling, parent,
		      (offset1 + node->max + t->sibling_space),
		      offset2, depth);
  }
}

/* ---------------------- Node Operations ---------------------------- */

void Willow_Create_Node(Willow_Tree_Type tree, 
			int x1, int y1, int x2, int y2,
			char *widget_id, Willow_Node *n,
			void *additional)
{
  *n = (Willow_Node)malloc(sizeof(struct Willow_Node_Struct));

  (*n)->height = abs(y2 - y1);
  (*n)->width  = abs(x2 - x1);
  (*n)->widget_id = widget_id;
  (*n)->connector_id = '\0';
  (*n)->current_x = x1;
  (*n)->current_y = y1;
  (*n)->sibling = NULL;
  (*n)->children = NULL;
  (*n)->last_child = NULL;
  (*n)->prev_sibling = NULL;
  (*n)->parent = NULL;
  (*n)->additional = additional;
}

void Willow_Insert_Node(Willow_Tree_Type t, Willow_Node n,
			Willow_Node parent)
{
  /* we need to keep track of the parent for deletion of the node */
  n->parent = parent;

  /* if parent==0 then we're inserting a root node */
  if (parent) {
    if (parent->last_child) {
      /* the parent node already has children */
      n->prev_sibling = parent->last_child;
      parent->last_child->sibling = n;
    } else {
      parent->children = n;
    }
    parent->last_child = n;
  } else {
    /* inserting the root node */
    t->root_node = n;
  }
}

void Willow_Delete_Node(Willow_Tree_Type t, Willow_Node n)
{
  char tcl_command[1024];

  /* Delete the graphic */
  sprintf(tcl_command, "%s delete %s",
	  t->canvas_name, n->widget_id);
  Tcl_Eval(t->interp, tcl_command);

  /* Delete the line connecting the node from the child to the parent */
  sprintf(tcl_command, "Willow_Delete_Connector %s %s",
	  t->canvas_name, n->connector_id);
  Tcl_Eval(t->interp, tcl_command);

  /* the only time that a parents data structure needs to be updated
     is when the child is the first or last child */

  /* Delete the node from Willow's internal tree structure */
  if (n->prev_sibling) {
    /* node has a previous sibling */
    n->prev_sibling->sibling = n->sibling;
  } else {
    /* update parent's first child record */
    n->parent->children = n->sibling;
  }
  
  if (n->sibling) {
    /* node has a next sibling */
    n->sibling->prev_sibling = n->prev_sibling;
  } else {
    /* update parent's last_child record */
    n->parent->last_child = n->prev_sibling;
  }

  /* should free(n) here */
}

void Willow_Modify_Node(Willow_Tree_Type t, Willow_Node old_node,
			Willow_Node new_node)
{
  char tcl_command[1024];

  /* Delete the old graphic */
  sprintf(tcl_command, "%s delete %s",
	  t->canvas_name, old_node->widget_id);
  Tcl_Eval(t->interp, tcl_command);

  /* to modify a node, overwrite the associated data of the node
     that's already there -- but do not modify the associate
     parent/child data */
  old_node->height = new_node->height;
  old_node->width  = new_node->width;
  old_node->widget_id = new_node->widget_id;
  old_node->current_x = new_node->current_x;
  old_node->current_y = new_node->current_y;
}

/* ------------------------------------------------------------------ */

void Willow_Move_Node(Willow_Tree_Type t, Willow_Node n,
		      int delta_x, int delta_y)
{
  char tcl_command[1024];
  
  sprintf(tcl_command, "%s move %s %d %d",
	  t->canvas_name, n->widget_id, delta_x, delta_y);
  Tcl_Eval(t->interp, tcl_command);
  /*
    sprintf(tcl_command, "update");
    Tcl_Eval(t->interp, tcl_command);
  */
  n->current_x += delta_x;
  n->current_y += delta_y;
}

/* Node bounding dimension retruns the max of the width of a node and
   the subtree rooted at that node. */

int Node_Bounding_Dimension(Willow_Tree_Type t, Willow_Node n, int depth)
{
  int w;

  /* The dimension of a node is the maximum of the dimension of that
     node (ie. the width or the height depending on tree orientation)
     and the bounding dimension of the subtree. */
  n->children_dimension = Subtree_Dimension(t, n, depth+1);

  if (t->orientation == WILLOW_VERTICAL) { 
    /* tree is vertical */
    w = MAX(n->width, n->children_dimension);
    t->max_level_size[depth] = MAX(t->max_level_size[depth], n->height);
  } else {
    /* tree is horizontal */
    w = MAX(n->height, n->children_dimension);
    t->max_level_size[depth] = MAX(t->max_level_size[depth], n->width);
  }

  n->max = w;

  return (w);
}

/* Subtree_Dimension calculates the width (or height, depending on the
   orientation) of the tree below a particular node. It will return 0
   for a node which has no children. This function is only called by
   Node_Bounding_Dimension. */

int Subtree_Dimension(Willow_Tree_Type t, Willow_Node parent, int depth)
{
  int sum=0;
  Willow_Node n;

  n = parent->children;

  while (n) {
    sum += Node_Bounding_Dimension(t, n, depth) + t->sibling_space;
    n = n->sibling;
  }

  if (parent->children) sum -= t->sibling_space;

  return (sum);
}

/* ------------------------------------------------------------------ */

/* Willow_Draw_Connector draws a line from the bottom of a parent
   token to the top of a child token (where bottom and top are
   relative to the current gravity -- vertical or horizontal). */

void Willow_Draw_Connector(Willow_Tree_Type t, Willow_Node n,
			   Willow_Node parent, int dx, int dy)
{
  char tcl_command[1024];
  int parent_x, parent_y;
  int x, y;

  if (parent == NULL) {
    return;
  }

  if (t->orientation == WILLOW_VERTICAL) {
    /* vertical */
    parent_x = parent->current_x + (parent->width / 2);
    parent_y = parent->current_y + parent->height;
    x = n->current_x + (n->width / 2);
    y = n->current_y;
  } else {
    /* horizontal */
    parent_x = parent->current_x + parent->width;
    parent_y = parent->current_y + (parent->height / 2);
    x = n->current_x;
    y = n->current_y + (n->height / 2);
  }

  if (n->connector_id == '\0') {
    /* create the connector from the parent to the child */
    sprintf(tcl_command, "Willow_Connect_Nodes %s %d %d %d %d",
	    t->canvas_name, parent_x, parent_y, x, y);
    Tcl_Eval(t->interp, tcl_command);
    /* The result of Willow_Connect_Nodes should be the indentifier
       of the connector object */
    n->connector_id = strdup(t->interp->result);
  } else {
    /* move the connector (faster than deleting and recreating?) */
    sprintf(tcl_command, "Willow_Move_Connector %s %s %d %d %d %d",
	    t->canvas_name, n->connector_id, parent_x, parent_y, x, y);
    Tcl_Eval(t->interp, tcl_command);
  }
}

/* ----------------------------------------------------------------- */

/* Willow_Set_Child_Space changes the space between the bottom of a
   node and the top of it's child. */
void Willow_Set_Child_Space(Willow_Tree_Type tree, int space)
{
  tree->child_space = space;
}

/* Willow_Set_Child_Space changes the space between the right side of
   a node and the left side of it's next sibling. */
void Willow_Set_Sibling_Space(Willow_Tree_Type tree, int space)
{
  tree->sibling_space = space;
}
