/*-----------------------------------------------------------------------------

File:		billBoard.c

Programmer:	Richard Goodwin

Description:	TCA module that provides a graphics server.  It is used to
                illustrate how to register messages and handlers.

		Very little error checking is done, X may crash if you use 
                unreasonable parameter values.

		Message handlers for the following message are implemented:
		BB_CLEAR_MSG
		BB_FLASH_MSG
		BB_FILL_MSG
		BB_DRAW_SQUARE_MSG
		BB_DRAW_CIRCLE_MSG
		BB_DRAW_LINE_MSG
		BB_DRAW_LINES_MSG
		BB_CONNECT_DOTS_MSG

		This is a stand-alone module.  It does not take any 
		command-line options.

		To run this example:
		  1. central 2 -lmdsi
		  2. billBoard
		  3. artist <your name>

-----------------------------------------------------------------------------*/

#include "tca/libc.h"
#include "tca.h"
#include "tca/ezx.h"
#include "billBoard.h"

/* Global Definitions */
#define BILLBOARD_MODULE_NAME "billBoard"

/* Global Variables */
Window theWindow;

int theColour = C_RED;

/* Utility routines. */
static void setColour(char *colour)
{
  if (strstr(colour,"red")) {
    printf("Red Fill\n");
    theColour = C_RED;
  } else if (strstr(colour,"white")) {
    printf("White Fill\n");
    theColour = C_WHITE;
  } else if (strstr(colour,"blue")) {
    printf("Blue Fill\n");
    theColour = C_BLUE;
  } else if (strstr(colour,"green")) {
    printf("Green Fill\n");
    theColour = C_LAWNGREEN;
  } else if (strstr(colour,"yellow")) {
    printf("Yellow Fill\n");
    theColour = C_YELLOW;
  } else {
    printf("Unknown Colour\n");
  }
  fflush(stdout);
  EZX_SetColor(theColour);
}

static char *getColour(void)
{
  if (theColour == C_RED) {
    return "red";
  } else if (theColour == C_WHITE) {
    return "white";
  } else if (theColour == C_BLUE) {
    return "blue";
  } else if (theColour == C_LAWNGREEN) {
    return "green";
  } else if (theColour == C_YELLOW) {
    return "yellow";
  } else if (theColour == C_WHITE) {
    return "white";
  } else {
    return "unknown";
  }
}

/* Message Handlers */

static void bb_clearHnd(TCA_REF_PTR ref, void *empty)
{
  EZX_ClearWindow(theWindow); 
}

static void bb_flashHnd(TCA_REF_PTR ref, void *empty)
{
  EZX_ClearWindow(theWindow); 
  EZX_Flush();
  sleep(1);
  EZX_SetColor(C_RED);
  EZX_FillRectangle(theWindow, 0, 0, BB_WIDTH, BB_HEIGHT);
  EZX_Flush();
  sleep(1);
  EZX_SetColor(C_YELLOW);
  EZX_FillRectangle(theWindow, 0, 0, BB_WIDTH, BB_HEIGHT);
  EZX_Flush();
  sleep(1);
  EZX_ClearWindow(theWindow); 
  EZX_Flush();
}

static void bb_fillHnd(TCA_REF_PTR ref, char **colour)
{
  setColour(*colour);
  EZX_FillRectangle(theWindow, 0, 0, BB_WIDTH, BB_HEIGHT);
  tcaFreeData(BB_FILL_MSG,colour);
  EZX_Flush();
}

static void bb_drawSquareHnd(TCA_REF_PTR ref, SQUARE_PTR sqr)
{
  setColour(sqr->colour);
  
  EZX_FillRectangle(theWindow, sqr->left, sqr->top, sqr->size, sqr->size);
  tcaFreeData(BB_DRAW_SQUARE_MSG,sqr);
  EZX_Flush();
}

static void bb_drawCircleHnd(TCA_REF_PTR ref, CIRCLE_PTR crl)
{
  setColour(crl->colour);
  if (crl->filled)
    EZX_FillCircle(theWindow, crl->center.x, crl->center.y, crl->radius);
  else
    EZX_DrawCircle(theWindow, crl->center.x, crl->center.y, crl->radius);
  tcaFreeData(BB_DRAW_CIRCLE_MSG,crl);
  EZX_Flush();
}

static void bb_drawLineHnd(TCA_REF_PTR ref, LINE_PTR line)
{
  EZX_DrawLine(theWindow, (*line)[0].x, (*line)[0].y, 
	       (*line)[1].x, (*line)[1].y);
  tcaFreeData(BB_DRAW_LINE_MSG,line);
  EZX_Flush();
}

static void bb_drawLinesHnd(TCA_REF_PTR ref, LINES_PTR lines)
{
  int i;

  for ( i=0; i<lines->num_lines; i++) {
    tcaInform(BB_DRAW_LINE_MSG,&(lines->lines[i]));
  }
  tcaFreeData(BB_DRAW_LINES_MSG,lines);
  EZX_Flush();
}

static void bb_connectDotsHnd(TCA_REF_PTR ref, POINT_LIST *list)
{
  LINE_TYPE line;
  POINT_LIST next;
  
  if(*list != NULL) {
    line[1] = (*list)->point;
    next = (*list)->next;
    while (next != NULL) {
      line[0] = line[1];
      line[1] = next->point;
      tcaInform(BB_DRAW_LINE_MSG,&(line));
      next = next->next;
    }
  }
  tcaFreeData(BB_CONNECT_DOTS_MSG,list);
  EZX_Flush();
}

static void bb_colorQueryHnd(TCA_REF_PTR ref, void *dummy)
{
  char *currentColour;

  currentColour = getColour();
  tcaReply(ref, &currentColour);
}

static void registerAll(void)
{
  /* Register any named formats. */
  tcaRegisterNamedFormatter(BB_POINT_NAME,BB_POINT_NAME_FORMAT);

  /* Register messages and handlers. */
  tcaRegisterInform(BB_CLEAR_MSG, BB_CLEAR_MSG_FORMAT, bb_clearHnd);
  tcaRegisterInform(BB_FLASH_MSG, BB_FLASH_MSG_FORMAT, bb_flashHnd);
  tcaRegisterInform(BB_FILL_MSG, BB_FILL_MSG_FORMAT, bb_fillHnd);
  tcaRegisterInform(BB_DRAW_SQUARE_MSG, BB_DRAW_SQUARE_MSG_FORMAT, 
		    bb_drawSquareHnd);
  tcaRegisterInform(BB_DRAW_CIRCLE_MSG, BB_DRAW_CIRCLE_MSG_FORMAT, 
		    bb_drawCircleHnd);
  tcaRegisterInform(BB_DRAW_LINE_MSG, BB_DRAW_LINE_MSG_FORMAT, 
		    bb_drawLineHnd);
  tcaRegisterInform(BB_DRAW_LINES_MSG, BB_DRAW_LINES_MSG_FORMAT, 
		    bb_drawLinesHnd);
  tcaRegisterInform(BB_CONNECT_DOTS_MSG, BB_CONNECT_DOTS_MSG_FORMAT, 
		    bb_connectDotsHnd);

  tcaRegisterQuery(BB_COLOR_QUERY, BB_COLOR_QUERY_FORMAT, 
		   BB_COLOR_REPLY_FORMAT, bb_colorQueryHnd);
}

void main(int argc, char *argv[])
{

  /* Initialize X11. */
  EZX_InitGraphics(getenv("DISPLAY"),BILLBOARD_MODULE_NAME);

  /* Create the window */
  theWindow = EZX_MakeWindow(BILLBOARD_MODULE_NAME,
			     BB_WIDTH, BB_HEIGHT,
			     "+20+20");

  /* Set the initial color */
  EZX_SetColor(theColour);
  
  /* Connect to TCA */
  tcaConnectModule(BILLBOARD_MODULE_NAME, tcaServerMachine());

  /* Register messages and handlers */
  registerAll();
  
  /* Wait until the system has started. */
  tcaWaitUntilReady();

  /* Enter the main loop handling messages. */
  tcaModuleListen();
}

